@@ -1,7 +1,12 @@
 #!/usr/bin/env python
 '''camera control for ptgrey chameleon camera'''
 
+# todo:
+#    - add ability to lower score and get past images sent
+#    - check bandwidth2 actual usage
+
 import time, threading, sys, os, numpy, Queue, errno, cPickle, signal, struct, fcntl, select, cStringIO
+import functools
 try:
     import cv2.cv as cv
 except ImportError:
@@ -65,13 +70,14 @@ class ImagePacket:
 
 class ThumbPacket:
     '''a thumbnail region sent to the ground station'''
-    def __init__(self, frame_time, regions, thumb, frame_loss, xmit_queue, pos):
+    def __init__(self, frame_time, regions, thumb, frame_loss, xmit_queue, pos, highscore):
         self.frame_time = frame_time
         self.regions = regions
         self.thumb = thumb
         self.frame_loss = frame_loss
         self.xmit_queue = xmit_queue
         self.pos = pos
+        self.highscore = highscore
 
 class CommandPacket:
     '''a command to run on the plane'''
@@ -90,6 +96,16 @@ class ImageRequest:
         self.frame_time = frame_time
         self.fullres = fullres
 
+class HeartBeat:
+    '''generic heartbeat to keep bsend alive'''
+    def __init__(self):
+        pass
+
+class CameraMessage:
+    '''critical camera message'''
+    def __init__(self, msg):
+        self.msg = msg
+
 class ChangeCameraSetting:
     '''update a camera setting'''
     def __init__(self, name, value):
@@ -101,6 +117,11 @@ class ChangeImageSetting:
     def __init__(self, name, value):
         self.name = name
         self.value = value
+
+class BlockCancel:
+    '''cancel object for callback on send1 complete'''
+    def __init__(self, blockid):
+        self.blockid = blockid
         
 
 class CameraModule(mp_module.MPModule):
@@ -117,6 +138,14 @@ class CameraModule(mp_module.MPModule):
         self.scan_thread2_h = None
         self.transmit_thread_h = None
         self.view_thread_h = None
+        self.airstart_thread_h = None
+        self.flying = True
+        self.terrain_alt = None
+        self.last_camparms = None
+
+        # prevent loopback of messages
+        for mtype in ['DATA16', 'DATA32', 'DATA64', 'DATA96']:
+            self.module('link').no_fwd_types.add(mtype)
 
         from MAVProxy.modules.lib.mp_settings import MPSettings, MPSetting
         self.camera_settings = MPSettings(
@@ -125,17 +154,25 @@ class CameraModule(mp_module.MPModule):
               MPSetting('capture_brightness', int, 150, 'Capture Brightness', range=(10, 300), increment=1),
               MPSetting('gamma', int, 950, 'Capture Gamma', range=(0,1000), increment=1),
               MPSetting('roll_stabilised', bool, True, 'Roll Stabilised'),
+              MPSetting('roll_limit', float, 30, 'Roll stabilisation limit'),
               MPSetting('altitude', int, 0, 'Altitude', range=(0,10000), increment=1),
-              MPSetting('filter_type', str, 'simple', 'Filter Type',
+              MPSetting('minalt', int, 30, 'MinAltitude', range=(0,10000), increment=1),
+              MPSetting('mpp100', float, 0.0977, 'MPPat100m', range=(0,10000), increment=0.001),
+              MPSetting('rotate180', bool, False, 'rotate180'),
+              MPSetting('camparms', str, None, 'camera parameters'),
+              MPSetting('filter_type', str, 'compactness', 'Filter Type',
                         choice=['simple', 'compactness']),
-              MPSetting('fullres', bool, False, 'Full Resolution'),
+              MPSetting('blue_emphasis', bool, False, 'BlueEmphasis'),
               MPSetting('framerate', str, 7, 'Frame Rate', choice=['1', '3', '7', '15']),
               MPSetting('process_divider', int, 1, 'Process Divider', range=(1,50), increment=1),
-              MPSetting('use_capture_time', bool, False, 'Use Capture Time'),
+              MPSetting('send2_divider', int, 1, 'Send2 Divider', range=(1,50), increment=1),
+              MPSetting('use_capture_time', bool, True, 'Use Capture Time'),
 
               MPSetting('gcs_address', str, None, 'GCS Address', tab='GCS'),
               MPSetting('gcs_view_port', int, 7543, 'GCS View Port', range=(1, 30000), increment=1),
               MPSetting('gcs_slave', str, None, 'GCS Slave'),
+              MPSetting('aircraft_address', str, None, 'Aircraft Address', tab='GCS'),
+              MPSetting('aircraft_port', int, 7544, 'Aircraft Port', range=(1, 30000), increment=1),
               
               MPSetting('bandwidth',  int, 40000, 'Link1 Bandwdith', 'Comms'),
               MPSetting('bandwidth2', int, 2000, 'Link2 Bandwidth'),
@@ -144,29 +181,32 @@ class CameraModule(mp_module.MPModule):
               MPSetting('send1', bool, True, 'Send on Link1'),
               MPSetting('send2', bool, True, 'Send on Link2'),
               MPSetting('maxqueue1', int, None, 'Maximum queue Link1'),
-              MPSetting('maxqueue2', int, 30, 'Maxqueue queue Link2'),
+              MPSetting('maxqueue2', int, 5, 'Maxqueue queue Link2'),
               MPSetting('thumbsize', int, 60, 'Thumbnail Size', range=(10, 200), increment=1),
               MPSetting('mosaic_thumbsize', int, 35, 'Mosaic Thumbnail Size', range=(10, 200), increment=1),
               MPSetting('use_bsend2', bool, True, 'Enable Link2'),
+              MPSetting('minspeed', int, 4, 'Min vehicle speed to save images'),
 
-              MPSetting('minscore', int, 75, 'Min Score Link1', range=(0,1000), increment=1, tab='Scoring'),
-              MPSetting('minscore2', int, 500, 'Min Score Link2', range=(0,1000), increment=1),
+              MPSetting('minscore', int, 500, 'Min Score Link1', range=(0,5000), increment=1, tab='Scoring'),
+              MPSetting('minscore2', int, 1200, 'Min Score Link2', range=(0,5000), increment=1),
               MPSetting('packet_loss', int, 0, 'Packet Loss', range=(0,100), increment=1, tab='Misc'),             
+              MPSetting('packet_loss2', int, 0, 'Link2 Packet Loss', range=(0,100), increment=1, tab='Misc'),             
               MPSetting('clock_sync', bool, False, 'GPS Clock Sync'),             
 
               MPSetting('brightness', float, 1.0, 'Display Brightness', range=(0.1, 10), increment=0.1,
-                        digits=2, tab='Display')
+                        digits=2, tab='Display'),
+              MPSetting('debug', bool, False, 'debug enable'),             
               ],
             title='Camera Settings'
             )
 
         self.image_settings = MPSettings(
-            [ MPSetting('MinRegionArea', float, 0.15, range=(0,100), increment=0.05, digits=2, tab='Image Processing'),
-              MPSetting('MaxRegionArea', float, 2.0, range=(0,100), increment=0.1, digits=1),
+            [ MPSetting('MinRegionArea', float, 0.3, range=(0,100), increment=0.05, digits=2, tab='Image Processing'),
+              MPSetting('MaxRegionArea', float, 4.0, range=(0,100), increment=0.1, digits=1),
               MPSetting('MinRegionSize', float, 0.1, range=(0,100), increment=0.05, digits=2),
-              MPSetting('MaxRegionSize', float, 2, range=(0,100), increment=0.1, digits=1),
+              MPSetting('MaxRegionSize', float, 3.0, range=(0,100), increment=0.1, digits=1),
               MPSetting('MaxRarityPct',  float, 0.02, range=(0,100), increment=0.01, digits=2),
-              MPSetting('RegionMergeSize', float, 3.0, range=(0,100), increment=0.1, digits=1),
+              MPSetting('RegionMergeSize', float, 1.0, range=(0,100), increment=0.1, digits=1),
               MPSetting('SaveIntermediate', bool, False)
               ],
             title='Image Settings')
@@ -191,6 +231,7 @@ class CameraModule(mp_module.MPModule):
         self.xmit_queue = 0
         self.xmit_queue2 = 0
         self.efficiency = 1.0
+        self.efficiency2 = 1.0
 
         self.last_watch = 0
         self.frame_loss = 0
@@ -198,30 +239,33 @@ class CameraModule(mp_module.MPModule):
         self.boundary_polygon = None
 
         self.bandwidth_used = 0
+        self.bandwidth_used2 = 0
         self.rtt_estimate = 0
+        self.rtt_estimate2 = 0
         self.bsocket = None
         self.bsend = None
         self.bsend2 = None
+        self.bsend2_thumb_total = 0
         self.bsend_slave = None
         self.framerate = 0
+        self.last_minscore = None
+        self.last_minscore2 = None
+        self.last_heartbeat = time.time()
+        self.last_heartbeat2 = time.time()
         
         # setup directory for images
-        self.camera_dir = os.path.join(self.logdir, "camera")
+        if self.logdir is None:
+            self.camera_dir = "camera"
+        else:
+            self.camera_dir = os.path.join(self.logdir, "camera")
         cuav_util.mkdir_p(self.camera_dir)
 
         self.mpos = mav_position.MavInterpolator(backlog=5000, gps_lag=0.3)
         self.joelog = cuav_joe.JoeLog(os.path.join(self.camera_dir, 'joe.log'), append=self.continue_mode)
-        # load camera params
-        path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..', '..',
-                            'cuav', 'data', 'chameleon1_arecont0.json')
-        if os.path.exists(path):
-            self.c_params.load(path)
-        else:
-            print("Warning: %s not found" % path)
 
         self.add_command('camera', self.cmd_camera,
                          'camera control',
-                         ['<start|stop|status|view|noview|boundary>',
+                         ['<start|stop|status|view|noview|boundary|airstart>',
                           'set (CAMERASETTING)'])
         self.add_completion_function('(CAMERASETTING)', self.settings.completion)
         self.add_command('remote', self.cmd_remote, "remote command", ['(COMMAND)'])
@@ -250,23 +294,27 @@ class CameraModule(mp_module.MPModule):
             self.running = False
             print("stopped camera capture")
         elif args[0] == "status":
-            print("Cap imgs:%u err:%u scan:%u fr:%.3f regions:%u jsize:%.0f xmitq:%u/%u lst:%u sq:%.1f eff:%.2f" % (
+            print("Cap imgs:%u err:%u scan:%u fr:%.3f regions:%u jsize:%.0f xmitq:%u/%u lst:%u sq:%.1f eff:%.2f/%.2f" % (
                 self.capture_count, self.error_count, self.scan_count, 
                 self.framerate,
                 self.region_count, 
                 self.jpeg_size,
-                self.xmit_queue, self.xmit_queue2, self.frame_loss, self.scan_queue.qsize(), self.efficiency))
+                self.xmit_queue, self.xmit_queue2, self.frame_loss, self.scan_queue.qsize(),
+                self.efficiency, self.efficiency2))
             print("self.bsend2 is ", self.bsend2)
             if self.bsend2 is not None:
-                self.bsend2.report(detailed=True)
+                self.bsend2.report(detailed=False)
         elif args[0] == "queue":
-            print("scan %u  save %u  transmit %u  eff %.1f  bw %.1f  rtt %.1f" % (
+            print("scan %u  save %u  transmit %u  eff %.1f/%.1f  bw %.1f/%.1f  rtt %.1f/%.1f" % (
                 self.scan_queue.qsize(),
                 self.save_queue.qsize(),
                 self.transmit_queue.qsize(),
                 self.efficiency,
+                self.efficiency2,
                 self.bandwidth_used,
-                self.rtt_estimate))
+                self.bandwidth_used2,
+                self.rtt_estimate,
+                self.rtt_estimate2))
         elif args[0] == "view":
             if self.mpstate.map is None:
                 print("Please load map module first")
@@ -292,7 +340,11 @@ class CameraModule(mp_module.MPModule):
                 self.boundary_polygon = cuav_util.polygon_load(self.boundary)
                 if self.mpstate.map is not None:
                     self.mpstate.map.add_object(mp_slipmap.SlipPolygon('boundary', self.boundary_polygon,
-                                                                       layer=1, linewidth=2, colour=(0,0,255)))                
+                                                                       layer=1, linewidth=2, colour=(0,0,255)))
+        elif args[0] == "airstart":
+            self.start_aircraft_bsend()
+            if self.airstart_thread_h is None:
+                self.airstart_thread_h = self.start_thread(self.airstart_thread)
         else:
             print(usage)
 
@@ -311,6 +363,7 @@ class CameraModule(mp_module.MPModule):
         error_count = 0
 
         print('Opening camera')
+        time.sleep(0.5)
         h = chameleon.open(1, self.camera_settings.depth, self.camera_settings.capture_brightness)
 
         print('Getting camare base_time')
@@ -328,10 +381,22 @@ class CameraModule(mp_module.MPModule):
                 error_count = 0
                 print('re-opening camera')
                 chameleon.close(h)
+                time.sleep(0.5)
                 h = chameleon.open(1, self.camera_settings.depth, self.camera_settings.capture_brightness)
         print('base_time=%f' % base_time)
         return h, base_time, frame_time
 
+    def check_camera_parms(self):
+        '''check for change in camera parameters'''
+        if self.camera_settings.camparms is None or self.last_camparms == self.camera_settings.camparms:
+            return
+        self.last_camparms = self.camera_settings.camparms
+        if os.path.exists(self.camera_settings.camparms):
+            self.c_params.load(self.camera_settings.camparms)
+            print("Loaded %s" % self.camera_settings.camparms)
+        else:
+            print("Warning: %s not found" % self.camera_settings.camparms)
+
     def capture_thread(self):
         '''camera capture thread'''
         print("running capture_thread")
@@ -350,6 +415,8 @@ class CameraModule(mp_module.MPModule):
             mode = 'w'
         gammalog = open(os.path.join(self.camera_dir, "gamma.log"), mode=mode)
 
+        last_successful_capture = None
+
         while not self.unload_event.wait(0.02):
             if not self.running:            
                 if h is not None:
@@ -358,13 +425,19 @@ class CameraModule(mp_module.MPModule):
                 continue
 
             try:
+                if h is not None and last_successful_capture is not None and time.time() - last_successful_capture > 5:
+                    self.send_message("Closing camera")
+                    print("Closing camera")
+                    chameleon.close(h)
+                    h = None
+                    last_successful_capture = None
+
                 if h is None:
                     h, base_time, last_frame_time = self.get_base_time()
                     last_capture_frame_time = last_frame_time
                     # put into continuous mode
                     chameleon.trigger(h, True)
 
-                capture_time = time.time()
                 if self.camera_settings.depth == 16:
                     im = numpy.zeros((960,1280),dtype='uint16')
                 else:
@@ -376,6 +449,10 @@ class CameraModule(mp_module.MPModule):
                     chameleon.set_framerate(h, int(self.camera_settings.framerate))
                     last_framerate = int(self.camera_settings.framerate)
 
+                self.check_camera_parms()
+
+                capture_time = time.time()
+                
                 # capture an image
                 frame_time, frame_counter, shutter = chameleon.capture(h, 1000, im)
                 if frame_time < last_capture_frame_time:
@@ -384,8 +461,12 @@ class CameraModule(mp_module.MPModule):
                 if last_frame_counter != 0:
                     self.frame_loss += frame_counter - (last_frame_counter+1)
 
+                if last_successful_capture is None:
+                    self.send_message("Camera OK")
+                last_successful_capture = time.time()
+
                 # discard based on process_divider setting
-                self.process_counter = (self.process_counter + 1) % self.camera_settings.process_divider
+                self.process_counter += 1
                 if self.process_counter % self.camera_settings.process_divider != 0:
                     continue
                 
@@ -428,7 +509,7 @@ class CameraModule(mp_module.MPModule):
             (frame_time,im) = self.save_queue.get()
             rawname = "raw%s" % cuav_util.frame_time(frame_time)
             frame_count += 1
-            if self.camera_settings.save_pgm != 0:
+            if self.camera_settings.save_pgm != 0 and self.flying:
                 if frame_count % self.camera_settings.save_pgm == 0:
                     chameleon.save_pgm('%s/%s.pgm' % (raw_dir, rawname), im)
 
@@ -447,17 +528,23 @@ class CameraModule(mp_module.MPModule):
             for name in self.image_settings.list():
                 scan_parms[name] = self.image_settings.get(name)
             scan_parms['SaveIntermediate'] = float(scan_parms['SaveIntermediate'])
+            scan_parms['BlueEmphasis'] = float(self.camera_settings.blue_emphasis)
+
+            if self.terrain_alt is not None:
+                altitude = self.terrain_alt
+                if altitude < self.camera_settings.minalt:
+                    altitude = self.camera_settings.minalt
+                scan_parms['MetersPerPixel'] = self.camera_settings.mpp100 * altitude / 100.0
             
             t1 = time.time()
             im_full = numpy.zeros((960,1280,3),dtype='uint8')
             im_640 = numpy.zeros((480,640,3),dtype='uint8')
             scanner.debayer(im, im_full)
+            if self.camera_settings.rotate180:
+                scanner.rotate180(im_full)
             scanner.downsample(im_full, im_640)
-            if self.camera_settings.fullres:
-                img_scan = im_full
-            else:
-                img_scan = im_640
-            regions = scanner.scan(img_scan)
+            img_scan = im_full
+            regions = scanner.scan(img_scan, scan_parms)
             if self.camera_settings.filter_type=='compactness':
                 calculate_compactness = True
             else:
@@ -481,7 +568,7 @@ class CameraModule(mp_module.MPModule):
     def get_plane_position(self, frame_time,roll=None):
         '''get a MavPosition object for the planes position if possible'''
         try:
-            pos = self.mpos.position(frame_time, 0,roll=roll)
+            pos = self.mpos.position(frame_time, 0, roll=roll, maxroll=self.camera_settings.roll_limit)
             return pos
         except mav_position.MavInterpolatorException as e:
             print str(e)
@@ -497,9 +584,57 @@ class CameraModule(mp_module.MPModule):
                                        thumb_filename, altitude=altitude)
 
 
+    def add_all_thumbs(self, pkt):
+        '''add to all_thumbs list'''
+        if pkt.pos is None or pkt.pos.altitude > 20:
+            # don't save ground photos
+            return
+
+    def check_send_newscore(self):
+        '''check if requested scores have changed, and send missing thumbs if needed'''
+        if self.last_minscore is None:
+            self.last_minscore = self.camera_settings.minscore
+        if self.last_minscore2 is None:
+            self.last_minscore2 = self.camera_settings.minscore2
+
+    def send_heartbeats(self):
+        '''possibly send heartbeat msgs'''
+        now = time.time()
+        if now - self.last_heartbeat > 5 and self.bsend.sendq_size() == 0:
+            self.last_heartbeat = now
+            self.send_heartbeat(self.bsend)
+        if (now - self.last_heartbeat2 > 5 and
+            self.camera_settings.use_bsend2 and
+            self.bsend2 is not None and
+            self.bsend2.sendq_size() == 0):
+            self.last_heartbeat2 = now
+            self.send_heartbeat(self.bsend2)
+
+    def airstart_thread(self):
+        '''thread for commands in aircraft when camera not running'''
+
+        while not self.unload_event.wait(0.02):
+            if self.transmit_thread_h is not None:
+                time.sleep(1)
+                continue
+            self.bsend.tick(packet_count=1000, max_queue=self.camera_settings.maxqueue1)
+            self.bsend2.tick(packet_count=1000, max_queue=self.camera_settings.maxqueue2)
+            self.check_commands(self.bsend)
+            self.check_commands(self.bsend2)
+            self.send_heartbeats()
+
+    def send_complete(self, blk_cancel):
+        '''called on complete of a send on first link'''
+        if blk_cancel.blockid is None or self.bsend2 is None:
+            # not sent on link2
+            return
+        # cancel link2 send of this block
+        self.bsend2.cancel(blk_cancel.blockid)
+
     def transmit_thread(self):
         '''thread for image transmit to GCS'''
         tx_count = 0
+        reg_count = 0
         skip_count = 0
         self.start_aircraft_bsend()
 
@@ -508,7 +643,9 @@ class CameraModule(mp_module.MPModule):
             self.bsend2.tick(packet_count=1000, max_queue=self.camera_settings.maxqueue2)
             self.check_commands(self.bsend)
             self.check_commands(self.bsend2)
+            self.send_heartbeats()
             if self.transmit_queue.empty():
+                self.check_send_newscore()
                 continue
 
             (frame_time, regions, im_full, im_640) = self.transmit_queue.get()
@@ -532,10 +669,15 @@ class CameraModule(mp_module.MPModule):
             self.efficiency = self.bsend.get_efficiency()
             self.bandwidth_used = self.bsend.get_bandwidth_used()
             self.rtt_estimate = self.bsend.get_rtt_estimate()
+            if self.bsend2 is not None:
+                self.efficiency2 = self.bsend2.get_efficiency()
+                self.rtt_estimate2 = self.bsend2.get_rtt_estimate()
+                self.bandwidth_used2 = self.bsend2.get_bandwidth_used()
 
             jpeg = None
 
             if len(regions) > 0:
+                reg_count += 1
                 lowscore = 0
                 highscore = 0
                 for r in regions:
@@ -544,39 +686,37 @@ class CameraModule(mp_module.MPModule):
                 
                 if self.camera_settings.transmit:
                     # send a region message with thumbnails to the ground station
-                    thumb = None
-                    if self.camera_settings.send1:
-                        thumb_img = cuav_mosaic.CompositeThumbnail(cv.GetImage(cv.fromarray(im_full)),
-                                                                   regions,
-                                                                   thumb_size=self.camera_settings.thumbsize)
-                        thumb = scanner.jpeg_compress(numpy.ascontiguousarray(cv.GetMat(thumb_img)), self.camera_settings.quality)
-                        
-                        pkt = ThumbPacket(frame_time, regions, thumb, self.frame_loss, self.xmit_queue, pos)
-                        
-                        buf = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
+                    thumb_img = cuav_mosaic.CompositeThumbnail(cv.GetImage(cv.fromarray(im_full)),
+                                                               regions,
+                                                               thumb_size=self.camera_settings.thumbsize)
+                    thumb = scanner.jpeg_compress(numpy.ascontiguousarray(cv.GetMat(thumb_img)), self.camera_settings.quality)
+                    pkt = ThumbPacket(frame_time, regions, thumb, self.frame_loss, self.xmit_queue, pos, highscore)
+                    buf = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
+
+                    # keep all thumbs so we can send more on score change
+                    self.add_all_thumbs(pkt)
+
+                    blk_cancel = BlockCancel(None)
+
+                    if self.camera_settings.send1 and highscore >= self.camera_settings.minscore:
+                        # send on primary link
                         self.bsend.set_bandwidth(self.camera_settings.bandwidth)
                         self.bsend.set_packet_loss(self.camera_settings.packet_loss)
-                        # print("sending thumb len=%u" % len(buf))
-                        self.bsend.send(buf,
-                                        dest=(self.camera_settings.gcs_address, self.camera_settings.gcs_view_port),
-                                        priority=1)
+                        self.bsend.send(buf, priority=highscore, callback=functools.partial(self.send_complete, blk_cancel))
+                        pkt.sent1 = True
 
                     # also send thumbnails via 900MHz telemetry
-                    if self.camera_settings.send2 and highscore >= self.camera_settings.minscore2:
-                        if thumb is None or lowscore < self.camera_settings.minscore2:
-                            # remove some of the regions
-                            regions = cuav_region.filter_regions(im_full, regions, min_score=self.camera_settings.minscore2,
-                                                                 filter_type=self.camera_settings.filter_type)
-                            thumb_img = cuav_mosaic.CompositeThumbnail(cv.GetImage(cv.fromarray(im_full)),
-                                                                       regions,
-                                                                       thumb_size=self.camera_settings.thumbsize)
-                            thumb = scanner.jpeg_compress(numpy.ascontiguousarray(cv.GetMat(thumb_img)), self.camera_settings.quality)
-                            pkt = ThumbPacket(frame_time, regions, thumb, self.frame_loss, self.xmit_queue, pos)
-                            
-                            buf = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
+                    if (self.camera_settings.send2 and
+                        highscore >= self.camera_settings.minscore2 and
+                        reg_count % self.camera_settings.send2_divider == 0):
+                        # send on secondary link
                             self.bsend2.set_bandwidth(self.camera_settings.bandwidth2)
-                            # print("sending thumb2 len=%u" % len(buf))
-                            self.bsend2.send(buf, priority=highscore)
+                            self.bsend2.set_packet_loss(self.camera_settings.packet_loss2)
+                            blk_cancel.blockid = self.bsend2.send(buf, priority=highscore)
+                            if self.camera_settings.debug:
+                                self.bsend2_thumb_total += len(buf)
+                                print("sent thumb bsend2 highscore=%u len=%u total=%u" % (
+                                    highscore, len(buf), self.bsend2_thumb_total))
 
             # Base how many images we send on the send queue size
             send_frequency = self.xmit_queue // 3
@@ -586,23 +726,34 @@ class CameraModule(mp_module.MPModule):
                 skip_count += 1
                 continue
             tx_count += 1
-            self.send_image(im_640, frame_time, pos, 0)
+            self.send_image(im_640, frame_time, pos, 0, bsend=self.bsend)
 
-    def send_image(self, img, frame_time, pos, priority):
+    def best_bsend(self, who):
+        '''choose the best link to use'''
+        if self.camera_settings.use_bsend2 and not self.bsend.is_alive(20) and self.bsend2 is not None:
+            if self.camera_settings.debug:
+                print("using bsend2 for %s" % who)
+            self.bsend2.set_packet_loss(self.camera_settings.packet_loss2)
+            self.bsend2.set_bandwidth(self.camera_settings.bandwidth2)
+            return self.bsend2
+        self.bsend.set_packet_loss(self.camera_settings.packet_loss)
+        self.bsend.set_bandwidth(self.camera_settings.bandwidth)
+        return self.bsend
+
+    def send_image(self, img, frame_time, pos, priority, bsend=None):
         '''send an image to the GCS'''
         jpeg = scanner.jpeg_compress(img, self.camera_settings.quality)
 
         # keep filtered image size
         self.jpeg_size = 0.95 * self.jpeg_size + 0.05 * len(jpeg)        
 
-        self.bsend.set_packet_loss(self.camera_settings.packet_loss)
-        self.bsend.set_bandwidth(self.camera_settings.bandwidth)
+        if bsend is None:
+            bsend = self.best_bsend('send_image')
+
         pkt = ImagePacket(frame_time, jpeg, self.xmit_queue, pos, priority)
         str = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
         # print("sending image len=%u" % len(str))
-        self.bsend.send(str,
-                        dest=(self.camera_settings.gcs_address, self.camera_settings.gcs_view_port),
-                        priority=priority)
+        bsend.send(str, priority=priority)
 
 
     def reload_mosaic(self, mosaic):
@@ -638,18 +789,24 @@ class CameraModule(mp_module.MPModule):
     def start_aircraft_bsend(self):
         '''start bsend for aircraft side'''
         if self.bsend is None:
-            self.bsend = block_xmit.BlockSender(0, bandwidth=self.camera_settings.bandwidth, debug=False)
+            self.bsend = block_xmit.BlockSender(self.camera_settings.aircraft_port,
+                                                bandwidth=self.camera_settings.bandwidth, debug=False,
+                                                dest_ip=self.camera_settings.gcs_address,
+                                                dest_port=self.camera_settings.gcs_view_port)
         if self.bsend2 is None:
             self.bsocket = MavSocket(self.mpstate.mav_master[0])
             self.bsend2 = block_xmit.BlockSender(mss=96, sock=self.bsocket, dest_ip='mavlink', dest_port=0, backlog=5, debug=False)
             self.bsend2.set_bandwidth(self.camera_settings.bandwidth2)
-        
+        self.send_heartbeats()
 
     def start_gcs_bsend(self):
         '''start up block senders for GCS side'''
         if self.bsend is None:
             self.bsend = block_xmit.BlockSender(self.camera_settings.gcs_view_port,
-                                                bandwidth=self.camera_settings.bandwidth)
+                                                bandwidth=self.camera_settings.bandwidth,
+                                                dest_ip=self.camera_settings.aircraft_address,
+                                                dest_port=self.camera_settings.aircraft_port)
+
         if self.bsend2 is None:
             self.bsocket = MavSocket(self.mpstate.mav_master[0])
             self.bsend2 = block_xmit.BlockSender(mss=96, sock=self.bsocket, dest_ip='mavlink',
@@ -740,11 +897,11 @@ class CameraModule(mp_module.MPModule):
 
             if self.camera_settings.gcs_slave is not None:
                 if self.bsend_slave is None:
-                    self.bsend_slave = block_xmit.BlockSender(0, bandwidth=self.camera_settings.bandwidth*10, debug=False)
+                    self.bsend_slave = block_xmit.BlockSender(0, bandwidth=self.camera_settings.bandwidth*10, debug=False,
+                                                              dest_ip=self.camera_settings.gcs_slave,
+                                                              dest_port=self.camera_settings.gcs_view_port)
                 # print("send bsend_slave")
-                self.bsend_slave.send(buf,
-                                      dest=(self.camera_settings.gcs_slave, self.camera_settings.gcs_view_port),
-                                      priority=1)
+                self.bsend_slave.send(buf, priority=1)
 
             if isinstance(obj, ThumbPacket):
                 # we've received a set of thumbnails from the plane for a positive hit
@@ -758,6 +915,8 @@ class CameraModule(mp_module.MPModule):
                 thumb_filename = '%s/v%s.jpg' % (thumb_dir, cuav_util.frame_time(obj.frame_time))
                 chameleon.save_file(thumb_filename, obj.thumb)
                 composite = cv.LoadImage(thumb_filename)
+                if composite is None:
+                    continue
                 thumbs = cuav_mosaic.ExtractThumbs(composite, len(obj.regions))
 
                 # log the joe positions
@@ -774,23 +933,25 @@ class CameraModule(mp_module.MPModule):
                 self.xmit_queue = obj.xmit_queue
                 thumb_count += 1
             
-                self.console.set_status('Lost', 'Lost %u' % self.frame_loss)
-                self.console.set_status('Regions', 'Regions %u' % region_count)
-                self.console.set_status('XMITQ', 'XMITQ %.0f' % self.xmit_queue)
-                self.console.set_status('Thumbs', 'Thumbs %u' % thumb_count)
-                self.console.set_status('ThumbSize', 'ThumbSize %.0f' % (thumb_total_bytes/thumb_count))
+                self.console.set_status('Lost', 'Lost %u' % self.frame_loss, row=6)
+                self.console.set_status('Regions', 'Regions %u' % region_count, row=6)
+                self.console.set_status('XMITQ', 'XMITQ %.0f' % self.xmit_queue, row=6)
+                self.console.set_status('Thumbs', 'Thumbs %u' % thumb_count, row=7)
+                self.console.set_status('ThumbSize', 'ThumbSize %.0f' % (thumb_total_bytes/thumb_count), row=7)
 
             if isinstance(obj, ImagePacket):
                 # we have an image from the plane
                 image_total_bytes += len(buf)
 
                 self.xmit_queue = obj.xmit_queue
-                self.console.set_status('XMITQ', 'XMITQ %.0f' % self.xmit_queue)
+                self.console.set_status('XMITQ', 'XMITQ %.0f' % self.xmit_queue, row=6)
 
                 # save it to disk
                 filename = '%s/v%s.jpg' % (view_dir, cuav_util.frame_time(obj.frame_time))
                 chameleon.save_file(filename, obj.jpeg)
                 img = cv.LoadImage(filename)
+                if img is None:
+                    continue
                 if img.width == 1280:
                     display_img = cv.CreateImage((640, 480), 8, 3)
                     cv.Resize(img, display_img)
@@ -815,9 +976,9 @@ class CameraModule(mp_module.MPModule):
                 image_count += 1
                 jpeg_total_bytes += len(obj.jpeg)
                 self.jpeg_size = 0.95 * self.jpeg_size + 0.05 * len(obj.jpeg)
-                self.console.set_status('Images', 'Images %u' % image_count)
-                self.console.set_status('JPGSize', 'JPG Size %.0f' % (jpeg_total_bytes/image_count))
-                self.console.set_status('ImageSize', 'ImageSize %.0f' % (image_total_bytes/image_count))
+                self.console.set_status('Images', 'Images %u' % image_count, row=6)
+                self.console.set_status('JPGSize', 'JPG Size %.0f' % (jpeg_total_bytes/image_count), row=6)
+                self.console.set_status('ImageSize', 'ImageSize %.0f' % (image_total_bytes/image_count), row=7)
 
             if isinstance(obj, CommandPacket):
                 self.handle_command_packet(obj, bsend)
@@ -825,6 +986,9 @@ class CameraModule(mp_module.MPModule):
             if isinstance(obj, CommandResponse):
                 print('REMOTE: %s' % obj.response)
 
+            if isinstance(obj, CameraMessage):
+                self.say(obj.msg)
+
     def start_thread(self, fn):
         '''start a thread running'''
         t = threading.Thread(target=fn)
@@ -899,6 +1063,10 @@ class CameraModule(mp_module.MPModule):
         if m.get_type() == 'SYSTEM_TIME' and self.camera_settings.clock_sync and self.capture_thread_h is not None:
             # optionally sync system clock on the capture side
             self.sync_gps_clock(m.time_unix_usec)
+        if m.get_type() == 'VFR_HUD':
+            self.flying = m.airspeed > self.camera_settings.minspeed or m.groundspeed > self.camera_settings.minspeed
+        if m.get_type() == "TERRAIN_REPORT":
+            self.terrain_alt = m.current_height
 
     def sync_gps_clock(self, time_usec):
         '''sync system clock with GPS time'''
@@ -927,17 +1095,7 @@ class CameraModule(mp_module.MPModule):
             buf = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
             self.start_gcs_bsend()
             print("Requesting image %s" % frame_time)
-            if self.camera_settings.use_bsend2:
-                if self.bsend2 is None:
-                    print("bsend2 not initialised")
-                    return
-                self.bsend2.set_bandwidth(self.camera_settings.bandwidth2)
-                self.bsend2.send(buf, priority=10000)
-            else:
-                if self.bsend is None:
-                    print("bsend not initialised")
-                    return
-                self.bsend.send(buf, priority=10000)
+            self.best_bsend('check_requested_images').send(buf, priority=10000)
 
     def handle_image_request(self, obj, bsend):
         '''handle ImageRequest from GCS'''
@@ -948,7 +1106,7 @@ class CameraModule(mp_module.MPModule):
             print("No file: %s" % filename)
             return
         try:
-            img = cuav_util.LoadImage(filename)
+            img = cuav_util.LoadImage(filename, rotate180=self.camera_settings.rotate180)
             img = numpy.asarray(cv.GetMat(img))
         except Exception:
             return
@@ -959,21 +1117,13 @@ class CameraModule(mp_module.MPModule):
         print("Sending image %s" % filename)
         self.send_image(img, obj.frame_time, None, 10000)
 
-    def send_packet(self, pkt):
+    def send_packet(self, pkt, bsend=None):
         '''send a packet from GCS'''
         buf = cPickle.dumps(pkt, cPickle.HIGHEST_PROTOCOL)
         self.start_gcs_bsend()
-        if self.camera_settings.use_bsend2:
-            if self.bsend2 is None:
-                print("bsend2 not initialised")
-                return
-            self.bsend2.set_bandwidth(self.camera_settings.bandwidth2)
-            self.bsend2.send(buf, priority=10000)
-        else:
-            if self.bsend is None:
-                print("bsend not initialised")
-                return
-            self.bsend.send(buf, priority=10000)
+        if bsend is None:
+            bsend = self.best_bsend('send_packet')
+        bsend.send(buf, priority=10000)
 
     def camera_settings_callback(self, setting):
         '''called on a changed camera setting'''
@@ -985,6 +1135,18 @@ class CameraModule(mp_module.MPModule):
         pkt = ChangeImageSetting(setting.name, setting.value)
         self.send_packet(pkt)
 
+    def send_heartbeat(self, bsend):
+        '''send a heartbeat'''
+        pkt = HeartBeat()
+        self.send_packet(pkt, bsend)
+
+    def send_message(self, msg, bsend=None):
+        '''send a message'''
+        if bsend is None:
+            bsend = self.best_bsend('send_message')
+        pkt = CameraMessage(msg)
+        self.send_packet(pkt, bsend)
+
 def init(mpstate):
     '''initialise module'''
     return CameraModule(mpstate)
